/*!
 * \file container/deque.h
 * \brief
 *      A statically allocated deque based on a ring buffer.
 *
 * \copyright Copyright (C) 2021 Christos Choutouridis <christos@choutouridis.net>
 *
 * <dl class=\"section copyright\"><dt>License</dt><dd>
 * The MIT License (MIT)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * </dd></dl>
 */

#ifndef utl_container_deque_h__
#define utl_container_deque_h__

#include <utl/core/impl.h>
#include <utl/container/ring_iterator.h>
#include <utl/container/range.h>
#include <array>
#include <atomic>

namespace utl {

/*!
 * \class deque
 * \brief
 *      A statically allocated deque based on a ring buffer
 *
 * The deque uses two ring_iterators one for the front and one for the rear. The iterators
 * are pointing to the next available spot, not on the last inserted spot. This way at the
 * initialization the iterators wont "pretend" to point to a valid item .
 *
 * We use a ring buffer of size \c N+1. We start the front iterator at the last location of the buffer
 * and the rear on the first. This way when the queue is full the iterators are pointing to the same location.
 *
 * \tparam Data_t       The char-like queued item type. Usually \c char
 * \tparam N            The size of deque
 * \tparam SemiAtomic   True for semi-atomic operation. In that case the \c ring_iterator is also atomic.
 * \note
 *      SemiAtomic means it is safe to access different ends from different threads. For example one thread can
 *      push only from front and another can pop from back to implement a queue.
 */
template <typename Data_t, size_t N, bool SemiAtomic =false>
class deque {
    public:
        // meta-identity type
        using type              = deque<Data_t, N>;
        using buffer_t          = std::array<Data_t, N+1>;  // We need N+1 spaces ring buffer for N spaces deque
        using iterator_t        = ring_iterator<Data_t*, N+1, SemiAtomic>;
        using range_t           = range<iterator_t>;

        // STL
        using value_type        = Data_t;
        using reference         = Data_t&;
        using const_reference   = const Data_t&;
        using pointer           = Data_t*;
        using const_pointer     = const Data_t*;
        using iterator          = iterator_t;
        using const_iterator    = const iterator_t;
        using reverse_iterator  = std::reverse_iterator<iterator>;
        using const_reverse_iterator = std::reverse_iterator<const_iterator>;

    //! \name Constructor / Destructor
    //! @{
    public:
        //! Default constructor
        constexpr deque () noexcept :
            data_{},
            f{data_.data(), N},
            r{data_.data()} {
                if constexpr (SemiAtomic)
                    std::atomic_thread_fence(std::memory_order_release);
            }

        //! fill contructor
        constexpr deque(const Data_t& value) noexcept {
            data_.fill(value);
            f = iterator(data_.data(), N);
            r = iterator(data_.data(), N);
            if constexpr (SemiAtomic)
                std::atomic_thread_fence(std::memory_order_release);
        }

        //! Initializer list contructor
        template <typename ...It>
        constexpr deque(It&& ...it) noexcept :
            data_{{std::forward<It>(it)...}},
            f(data_.data(), N),
            r(data_.data(), sizeof...(It)) {
                if constexpr (SemiAtomic)
                    std::atomic_thread_fence(std::memory_order_release);
            }

        deque(const deque&) = delete;                   //!< No copies
        deque& operator= (const deque&) = delete;       //!< No copy assignments
        ~deque () = default;                            //!< default destructor
    //! @}

    //! \name Iterators
    //! @{
    public:
        constexpr iterator begin() noexcept               { iterator ret = f; return ++ret; }
        constexpr const_iterator begin() const noexcept   { iterator ret = f; return ++ret; }
        constexpr const_iterator cbegin() const noexcept  { iterator ret = f; return ++ret; }

        constexpr iterator end() noexcept                 { return r; }
        constexpr const_iterator end() const noexcept     { return r; }
        constexpr const_iterator cend() const noexcept    { return r; }

        constexpr reverse_iterator rbegin() noexcept              { return r; }
        constexpr const_reverse_iterator rbegin() const noexcept  { return r; }
        constexpr const_reverse_iterator crbegin() const noexcept { return r; }

        constexpr reverse_iterator rend() noexcept                { reverse_iterator ret = f; return ++ret; }
        constexpr const_reverse_iterator rend() const noexcept    { reverse_iterator ret = f; return ++ret; }
        constexpr const_reverse_iterator crend() const noexcept   { reverse_iterator ret = f; return ++ret; }
    //! @}

    //! \name Capacity
    //! @{
    public:
        //! \return The size of the deque. The items currently in queue.
        constexpr size_t size() noexcept {
            return r - (f +1);
        }
        constexpr size_t size() const noexcept {
            return r - (f +1);
        }
        //! \return The maximum size of the deque. The items the queue can hold.
        constexpr size_t max_size() noexcept { return N; }
        //! \return The capacity of the deque. The items the queue can hold.
        constexpr size_t capacity() noexcept { return N; }
        //! \return True if the deque is empty
        constexpr bool   empty()    noexcept { return size() == 0 ? true : false; }
        //! \return True if the deque is full
        constexpr bool   full()     noexcept { return size() == N ? true : false; }

    //! @}

    //! \name Member access
    //! @{
    public:
        //! \brief  Clears-empty the deque and return it to init state, without
        //!         really deleting the contents.
        constexpr void clear() noexcept {
            f = iterator_t(data_.data(), N);
            r = iterator_t(data_.data());
            if constexpr (SemiAtomic)
                std::atomic_thread_fence(std::memory_order_release);
        }
        //! \brief      Push an item in the front of the deque
        //! \param it   The item to push
        constexpr void push_front (const Data_t& it) noexcept {
            if (full())     return;
            *f = it;
            --f;        // keep this separate for thread safety
        }
        //! \brief      Push an item in the back of the deque
        //! \param it   The item to push
        constexpr void push_back (const Data_t& it) noexcept {
            if (full())     return;
            *r = it;
            ++r;        // keep this separate for thread safety
        }
        //! \brief      Extract an item from the front of the deque and remove it from the deque
        //! \param it   The item to push
        constexpr Data_t pop_front () noexcept {
            if (empty())    return Data_t{};
            return *++f;
        }
        //! \brief      Extract an item from the back of the deque and remove it from the deque
        //! \param it   The item to push
        constexpr Data_t pop_back () noexcept {
            if (empty())    return Data_t{};
            return *--r;
        }

        //! \brief      Get a reference to the item in the front of the deque without extracting it.
        //! \return     Reference to the item
        constexpr Data_t& front() noexcept { iterator_t it = f; return *++it; }
        constexpr const Data_t& front() const noexcept { iterator_t it = f; return *++it; }

        //! \brief      Get a reference to the item in the front of the deque without extracting it.
        //! \return     Reference to the item
        constexpr Data_t& back() noexcept { iterator_t it = r; return *--it; }
        constexpr const Data_t& back() const noexcept { iterator_t it = r; return *--it; }

        //! \brief      Get a pointer to the begin of the items on the deque
        //! \return
        constexpr Data_t* data() noexcept { return &front(); }
        constexpr const Data_t* data() const noexcept { return &front(); }

        //! \brief      Get a range for the data in queue
        //! \return     A begin-end iterator pair struct
        constexpr range_t contents () noexcept { iterator_t b = f; return {++b, r}; }
        constexpr const range_t contents () const noexcept { iterator_t b = f; return {++b, r}; }

    //! @}
    private:
        buffer_t    data_{};            //!< The statically allocated buffer
        iterator_t  f{data_.data(), N}; //!< A ring iterator for the front (points to the next available location)
        iterator_t  r{data_.data()};    //!< A ring iterator for the rear (points to the next available location).
};

} // namespace utl

#endif /* utl_container_deque_h__ */
